import pytest

from sklego.transformers import RandomAdder
from tests.conftest import id_func


@pytest.mark.parametrize("transformer", [
    RandomAdder(),
], ids=id_func)
def test_same_shape_regression(transformer, random_xy_dataset_regr):
    X, y = random_xy_dataset_regr
    assert transformer.fit(X, y).transform(X).shape == X.shape


@pytest.mark.parametrize("transformer", [
    RandomAdder(),
], ids=id_func)
def test_shape_classification(transformer, random_xy_dataset_clf):
    X, y = random_xy_dataset_clf
    assert transformer.fit(X, y).transform(X).shape == X.shape



import numpy as np
import pandas as pd

from sklearn import clone
from sklearn.base import BaseEstimator
from sklearn.utils.validation import check_is_fitted, check_X_y, check_array

from sklego.common import as_list


class GroupedEstimator(BaseEstimator):
    """
    Construct an estimator per data group. Splits data by values of a
    single column and fits one estimator per such column.

    :param estimator: the model/pipeline to be applied per group
    :param groups: the column(s) of the matrix/dataframe to select as a grouping parameter set
    :param use_fallback: weather or not to fall back to a general model in case
    the group parameter is not found during `.predict()`
    """

    def __init__(self, estimator, groups, use_fallback=True):
        self.estimator = estimator
        self.groups = groups
        self.use_fallback = use_fallback

    def fit(self, X, y):
        """
        Fit the model using X, y as training data. Will also learn the groups
        that exist within the dataset.

        :param X: array-like, shape=(n_columns, n_samples,) training data.
        :param y: array-like, shape=(n_samples,) training data.
        :return: Returns an instance of self.
        """
        check_X_y(X, y)
        pred_col = 'the-column-that-i-want-to-predict-but-dont-have-the-name-for'
        if isinstance(X, np.ndarray):
            X = pd.DataFrame(X, columns=[str(_) for _ in range(X.shape[1])])
        X = X.assign(**{pred_col: y})

        self.group_colnames_ = [str(_) for _ in as_list(self.groups)]
        if any([c not in X.columns for c in self.group_colnames_]):
            raise ValueError(f"{self.group_colnames_} not in {X.columns}")
        self.X_colnames_ = [_ for _ in X.columns if _ not in self.group_colnames_ and _ is not pred_col]
        self.fallback_ = None
        if self.use_fallback:
            subset_x = X[self.X_colnames_]
            self.fallback_ = clone(self.estimator).fit(subset_x, y)

        self.groups_ = X[self.group_colnames_].drop_duplicates()

        self.estimators_ = (X
                            .groupby(self.group_colnames_)
                            .apply(lambda d: clone(self.estimator).fit(d[self.X_colnames_], d[pred_col]))
                            .to_dict())
        return self

    def predict(self, X):
        """
        Predict new data by making random guesses.

        :param X: array-like, shape=(n_columns, n_samples,) training data.
        :return: array, shape=(n_samples,) the predicted data
        """
        check_array(X)
        check_is_fitted(self, ['estimators_', 'groups_', 'group_colnames_',
                               'X_colnames_', 'fallback_'])
        if isinstance(X, np.ndarray):
            X = pd.DataFrame(X, columns=[str(_) for _ in range(X.shape[1])])

        if any([c not in X.columns for c in self.group_colnames_]):
            raise ValueError(f"group columns {self.group_colnames_} not in {X.columns}")
        if any([c not in X.columns for c in self.X_colnames_]):
            raise ValueError(f"columns to use {self.X_colnames_} not in {X.columns}")

        try:
            return (X
                    .groupby(self.group_colnames_, as_index=False)
                    .apply(lambda d: pd.DataFrame(
                        self.estimators_.get(d.name, self.fallback_).predict(d[self.X_colnames_]), index=d.index))
                    .values
                    .squeeze())
        except AttributeError:
            culprits = set(pd.concat([X[self.group_colnames_].drop_duplicates().assign(new=1),
                                      self.groups_.assign(new=0)])
                             .drop_duplicates()
                             .loc[lambda d: d['new'] == 1]
                             .itertuples())
            raise ValueError(f"found a group(s) {culprits} in `.predict` that was not in `.fit`")



import os
import pandas as pd


def load_chicken(give_pandas=False):
    """
    Loads the chicken dataset. The chicken data has 578 rows and 4 columns
    from an experiment on the effect of diet on early growth of chicks.
    The body weights of the chicks were measured at birth and every second
    day thereafter until day 20. They were also measured on day 21.
    There were four groups on chicks on different protein diets.

    :param give_pandas: give the pandas dataframe instead of X, y matrices (default=False)
    :return: (**X**, **y**) unless dataframe is returned

    :Example:

    >>> from sklego.datasets import load_chicken
    >>> X, y = load_chicken()
    >>> X.shape
    (578, 3)
    >>> y.shape
    (578,)
    >>> load_chicken(give_pandas=True).columns
    Index(['weight', 'time', 'chick', 'diet'], dtype='object')

    The datasets can be found in the folowing sources:

    - Crowder, M. and Hand, D. (1990), Analysis of Repeated Measures, Chapman and Hall (example 5.3)
    - Hand, D. and Crowder, M. (1996), Practical Longitudinal Data Analysis, Chapman and Hall (table A.2)
    """
    dirname = os.path.dirname(__file__)
    filepath = os.path.join(dirname, "data", "chickweight.csv")
    df = pd.read_csv(filepath)
    if give_pandas:
        return df
    return df[['time', 'diet', 'chick']].values, df['weight'].values



from sklego.datasets import load_chicken


def test_chickweight1():
    X, y = load_chicken()
    assert X.shape == (578, 3)
    assert y.shape[0] == 578


def test_chickweight2():
    df = load_chicken(give_pandas=True)
    assert df.shape == (578, 4)



from sklearn.base import BaseEstimator, TransformerMixin, MetaEstimatorMixin, clone
from sklearn.utils import check_array, check_X_y
from sklearn.utils.validation import FLOAT_DTYPES, check_random_state, check_is_fitted
from patsy import dmatrix, build_design_matrices, PatsyError
import numpy as np

from sklego.common import TrainOnlyTransformerMixin


class RandomAdder(TrainOnlyTransformerMixin, BaseEstimator):
    def __init__(self, noise=1, random_state=None):
        self.noise = noise
        self.random_state = random_state

    def fit(self, X, y):
        super().fit(X, y)
        X, y = check_X_y(X, y, estimator=self, dtype=FLOAT_DTYPES)
        self.dim_ = X.shape[1]

        return self

    def transform_train(self, X):
        rs = check_random_state(self.random_state)
        check_is_fitted(self, ['dim_'])

        X = check_array(X, estimator=self, dtype=FLOAT_DTYPES)

        return X + rs.normal(0, self.noise, size=X.shape)


class EstimatorTransformer(TransformerMixin, MetaEstimatorMixin, BaseEstimator):
    """
    Allows using an estimator such as a model as a transformer in an earlier step of a pipeline

    :param estimator: An instance of the estimator that should be used for the transformation
    :param predict_func: The function called on the estimator when transforming e.g. (`predict`, `predict_proba`)
    """

    def __init__(self, estimator, predict_func='predict'):
        self.estimator = estimator
        self.predict_func = predict_func

    def fit(self, X, y):
        """Fits the estimator"""
        X, y = check_X_y(X, y, estimator=self, dtype=FLOAT_DTYPES)

        self.estimator_ = clone(self.estimator)
        self.estimator_.fit(X, y)
        return self

    def transform(self, X):
        """
        Applies the `predict_func` on the fitted estimator.

        Returns an array of shape `(X.shape[0], )`.
        """
        check_is_fitted(self, 'estimator_')
        return getattr(self.estimator_, self.predict_func)(X)


class PatsyTransformer(TransformerMixin, BaseEstimator):
    """
    The patsy transformer offers a method to select the right columns
    from a dataframe as well as a DSL for transformations. It is inspired
    from R formulas.

    This is can be useful as a first step in the pipeline.

    :param formula: a patsy-compatible formula
    """

    def __init__(self, formula):
        self.formula = formula

    def fit(self, X, y=None):
        """Fits the estimator"""
        X_ = dmatrix(self.formula, X)
        assert np.array(X_).shape[0] == np.array(X).shape[0]
        self.design_info_ = X_.design_info
        return self

    def transform(self, X):
        """
        Applies the formula to the matrix/dataframe X.

        Returns an design array that can be used in sklearn pipelines.
        """
        check_is_fitted(self, 'design_info_')
        try:
            return build_design_matrices([self.design_info_], X)[0]
        except PatsyError as e:
            raise RuntimeError from e



from setuptools import setup, find_packages
import os

base_packages = ["numpy>=1.15.4", "scipy>=1.2.0", "scikit-learn>=0.20.2",
                 "pandas>=0.23.4", "patsy>=0.5.1", "autograd>=1.2"]

docs_packages = ["sphinx>=1.8.5", "sphinx_rtd_theme>=0.4.3"]
dev_packages = docs_packages + ["flake8>=3.6.0", "matplotlib>=3.0.2", "pytest>=4.0.2",
                                "nbval>=0.9.1", "plotnine>=0.5.1", "jupyter>=1.0.0",
                                "jupyterlab>=0.35.4"]


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name="scikit-lego",
    version="0.1.5",
    packages=find_packages(exclude=['notebooks']),
    long_description=read('readme.md'),
    long_description_content_type='text/markdown',
    install_requires=base_packages,
    extras_require={
        "docs": docs_packages,
        "dev": dev_packages
    }
)



import autograd.numpy as np
from autograd import grad
from autograd.test_util import check_grads

from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.utils import check_X_y
from sklearn.utils.validation import check_is_fitted, check_array, FLOAT_DTYPES


class DeadZoneRegressor(BaseEstimator, RegressorMixin):
    def __init__(self, threshold=0.3, relative=False, effect="linear", n_iter=2000, stepsize=0.01, check_grad=False):
        self.threshold = threshold
        self.relative = relative
        self.effect = effect
        self.n_iter = n_iter
        self.stepsize = stepsize
        self.check_grad = check_grad
        self.allowed_effects = ("linear", "quadratic", "constant")
        self.loss_log_ = None
        self.wts_log_ = None
        self.deriv_log_ = None
        self.coefs_ = None

    def fit(self, X, y):
        X, y = check_X_y(X, y, estimator=self, dtype=FLOAT_DTYPES)
        if self.effect not in self.allowed_effects:
            raise ValueError(f"effect {self.effect} must be in {self.allowed_effects}")

        def deadzone(errors):
            if self.effect == "linear":
                return np.where(errors > self.threshold, errors, np.zeros(errors.shape))
            if self.effect == "quadratic":
                return np.where(errors > self.threshold, errors**2, np.zeros(errors.shape))

        def training_loss(weights):
            diff = np.abs(np.dot(X, weights) - y)
            if self.relative:
                diff = diff / y
            return np.mean(deadzone(diff))

        n, k = X.shape

        # Build a function that returns gradients of training loss using autograd.
        training_gradient_fun = grad(training_loss)

        # Check the gradients numerically, just to be safe.
        weights = np.random.normal(0, 1, k)
        if self.check_grad:
            check_grads(training_loss, modes=['rev'])(weights)

        # Optimize weights using gradient descent.
        self.loss_log_ = np.zeros(self.n_iter)
        self.wts_log_ = np.zeros((self.n_iter, k))
        self.deriv_log_ = np.zeros((self.n_iter, k))
        for i in range(self.n_iter):
            weights -= training_gradient_fun(weights) * self.stepsize
            self.wts_log_[i, :] = weights.ravel()
            self.loss_log_[i] = training_loss(weights)
            self.deriv_log_[i, :] = training_gradient_fun(weights).ravel()
        self.coefs_ = weights
        return self

    def predict(self, X):
        X = check_array(X, estimator=self, dtype=FLOAT_DTYPES)
        check_is_fitted(self, ['coefs_'])
        return np.dot(X, self.coefs_)
